/*
 * Test directory reading operations.
 *
 * Copyright (C) 2020, 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/err.h>

#include <dirent.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <fsclient/client.h>
#include <systypes/fcntl.h>



#define DIRENT_CORE_SIZE (sizeof(struct dirent) - sizeof(((struct dirent *) 0)->d_name))



/* Return a directory entry. This must be freed by the caller after use. */

static struct dirent *read_directory_entry(file_t *file)
{
  char buffer[DIRENT_CORE_SIZE];
  offset_t nread = client_read(file, buffer, DIRENT_CORE_SIZE);

  /* Stop if no new structure can be successfully read. */

  if (nread != DIRENT_CORE_SIZE)
    return NULL;

  struct dirent *dirent = (struct dirent *) buffer;
  offset_t remaining = dirent->d_reclen - DIRENT_CORE_SIZE;

  /* Allocate a buffer for the complete structure. */

  char *entry = (char *) calloc(DIRENT_CORE_SIZE + remaining, sizeof(char));

  if (entry == NULL)
    return NULL;

  /* Copy the start of the entry into a new buffer. */

  memcpy(entry, buffer, DIRENT_CORE_SIZE);

  /* Append to the entry buffer. */

  char *current = entry + DIRENT_CORE_SIZE;

  nread = client_read(file, current, remaining);

  /* Stop if no complete structure can be successfully read. */

  if (nread != remaining)
  {
    free(entry);
    return NULL;
  }

  return (struct dirent *) entry;
}



static file_t *open_file(char *filename, bool have_uid, sys_uid_t uid)
{
  /* With a user, open a user-specific file opener. */

  if (have_uid)
  {
    l4_cap_idx_t opener = client_open_for_user((user_t) {uid, uid, 0022});

    if (l4_is_invalid_cap(opener))
    {
      printf("Could not obtain opener for file.\n");
      return NULL;
    }

    /* Invoke the open method to receive the file reference. */

    return client_open_using(filename, O_DIRECTORY, opener);
  }
  else
  {
    return client_open(filename, O_DIRECTORY);
  }
}



static long open_directory(file_t *file)
{
  /* Register the reader for notification. */

  return client_set_blocking(file, NOTIFY_CONTENT_AVAILABLE | NOTIFY_PEER_CLOSED);
}



int main(int argc, char *argv[])
{
  if (argc < 2)
  {
    printf("Need a directory name and an optional user identifier (if used with a filesystem).\n");
    return 1;
  }

  char *filename = argv[1];
  bool have_uid = (argc > 2) && strlen(argv[2]);
  sys_uid_t uid = have_uid ? atoi(argv[2]) : 0;
  file_t *file = open_file(filename, have_uid, uid);

  if (file == NULL)
  {
    printf("Could not obtain directory.\n");
    return 1;
  }

  long err = open_directory(file);

  if (err)
  {
    printf("Could not subscribe to notifications: %s\n", l4sys_errtostr(err));
    return 1;
  }

  struct dirent *dirent;

  while ((dirent = read_directory_entry(file)) != NULL)
  {
    printf("> %s\n", dirent->d_name);
    free(dirent);
  }

  printf("Directory shown.\n");

  /* Open again, reading a single entry only. */

  file = open_file(filename, have_uid, uid);

  if (file == NULL)
  {
    printf("Could not obtain directory.\n");
    return 1;
  }

  err = open_directory(file);

  if (err)
  {
    printf("Could not subscribe to notifications: %s\n", l4sys_errtostr(err));
    return 1;
  }

  dirent = read_directory_entry(file);

  if (dirent != NULL)
  {
    printf("> %s\n", dirent->d_name);
    free(dirent);
  }

  printf("Entry shown.\n");

  return 0;
}

// vim: tabstop=2 expandtab shiftwidth=2
